/*
 * IniFile.java
 * A class for handling Windows-style INI files.
 * 
 * Copyright Steve DeGroof, degroof@mindspring.com,
 * http://www.mindspring.com/~degroof
 * 
 * The source code provided here should be considered example
 * code. That is, you can use or modify it without permission. On the
 * other hand, you're using the code at your own risk.
 */

import java.io.*;
import java.util.*;

public class IniFile extends Object {

	protected Vector lines;
	protected Vector subjects;
	protected Vector variables;
	protected Vector values;
	protected String fileName;
	protected boolean saveOnChange = false;

	public IniFile(String name) {
		this(name, false);
	}

	public IniFile(String name, boolean save) {
		saveOnChange = save;
		fileName = name;
		if (!((new File(name)).exists())) {
			if (!createFile())
				return;
		}
		loadFile();
		parseLines();
	}

	public void loadFile() {
		lines = new Vector();
		subjects = new Vector();
		variables = new Vector();
		values = new Vector();

		try {
			DataInputStream ini = new DataInputStream(new BufferedInputStream(
					new FileInputStream(fileName)));
			String line = "";

			while (true) {
				line = ini.readLine();
				if (line == null)
					break;
				else
					lines.addElement(line.trim());
			}
			ini.close();
		} catch (IOException e) {
		}
	}

	protected boolean createFile() {
		try {
			DataOutputStream newFile = new DataOutputStream(
					new FileOutputStream(fileName));
			newFile.writeBytes(";INI File: " + fileName
					+ System.getProperty("line.separator"));
			newFile.close();
			return true;
		} catch (IOException e) {
			return false;
		}
	}

	protected void parseLines() {
		String currentLine = null; 
		String currentSubject = null; 
		for (int i = 0; i < lines.size(); i++)
		{
			currentLine = (String) lines.elementAt(i);
			if (isaSubject(currentLine))
			{
				currentSubject = currentLine.substring(1,
						currentLine.length() - 1);
			} else if (isanAssignment(currentLine))
			{
				String assignment = currentLine;
				addAssignment(currentSubject, assignment);
			}
		}
	}

	protected boolean addAssignment(String subject, String assignment) {
		String value;
		String variable;
		int index = assignment.indexOf("=");
		variable = assignment.substring(0, index);
		value = assignment.substring(index + 1, assignment.length());
		if ((value.length() == 0) || (variable.length() == 0))
			return false;
		else
			return addValue(subject, variable, value, false);
	}

	public boolean setValue(String subject, String variable, String value) {
		boolean result = addValue(subject, variable, value, true);
		if (saveOnChange)
			saveFile();
		return result;
	}

	protected boolean addValue(String subject, String variable, String value,
			boolean addToLines) {

		if ((subject == null) || (subject.length() == 0))
			return false;

		if ((variable == null) || (variable.length() == 0))
			return false;

		if (!subjects.contains(subject)) {
			subjects.addElement(subject);
			variables.addElement(new Vector());
			values.addElement(new Vector());
		}

		int subjectIndex = subjects.indexOf(subject);
		Vector subjectVariables = (Vector) (variables.elementAt(subjectIndex));
		Vector subjectValues = (Vector) (values.elementAt(subjectIndex));
		if (!subjectVariables.contains(variable)) {
			subjectVariables.addElement(variable);
			subjectValues.addElement(value);
		}
		int variableIndex = subjectVariables.indexOf(variable);
		subjectValues.setElementAt(value, variableIndex);

		if (addToLines)
			setLine(subject, variable, value);

		return true;
	}

	protected boolean isaSubject(String line) {
		return (line.startsWith("[") && line.endsWith("]"));
	}

	protected void setLine(String subject, String variable, String value) {

		int subjectLine = findSubjectLine(subject);
		if (subjectLine == -1) {
			addSubjectLine(subject);
			subjectLine = lines.size() - 1;
		}

		int endOfSubject = endOfSubject(subjectLine);

		int lineNumber = findAssignmentBetween(variable, subjectLine,
				endOfSubject);

		if (lineNumber == -1)
			lines.insertElementAt(variable + "=" + value, endOfSubject);
		else
			lines.setElementAt(variable + "=" + value, lineNumber);
	}

	protected int findAssignmentLine(String subject, String variable) {
		int start = findSubjectLine(subject);
		int end = endOfSubject(start);
		return findAssignmentBetween(variable, start, end);
	}

	protected int findAssignmentBetween(String variable, int start, int end) {
		for (int i = start; i < end; i++) {
			if (((String) lines.elementAt(i)).startsWith(variable + "="))
				return i;
		}
		return -1;
	}

	protected void addSubjectLine(String subject) {
		lines.addElement("[" + subject + "]");
	}

	protected int findSubjectLine(String subject) {
		String line;
		String formattedSubject = "[" + subject + "]";
		for (int i = 0; i < lines.size(); i++) {
			line = (String) lines.elementAt(i);
			if (formattedSubject.equals(line))
				return i;
		}
		return -1;
	}

	protected int endOfSubject(int start) {
		int endIndex = start + 1;
		if (start >= lines.size())
			return lines.size();
		for (int i = start + 1; i < lines.size(); i++) {
			if (isanAssignment((String) lines.elementAt(i)))
				endIndex = i + 1;
			if (isaSubject((String) lines.elementAt(i)))
				return endIndex;
		}
		return endIndex;
	}

	protected boolean isanAssignment(String line) {
		if ((line.indexOf("=") != -1) && (!line.startsWith(";")))
			return true;
		else
			return false;
	}

	public Vector getLines() {
		return (Vector) lines.clone();
	}

	public String[] getVariables(String subject) {
		String[] v;
		int index = subjects.indexOf(subject);
		if (index != -1) {
			Vector vars = (Vector) (variables.elementAt(index));
			v = new String[vars.size()];
			vars.copyInto(v);
			return v;
		} else {
			v = new String[0];
			return v;
		}
	}

	public String[] getSubjects() {
		String[] s = new String[subjects.size()];
		subjects.copyInto(s);
		return s;
	}

	public String getValue(String subject, String variable) {
		int subjectIndex = subjects.indexOf(subject);
		if (subjectIndex == -1)
			return "";
		Vector valVector = (Vector) (values.elementAt(subjectIndex));
		Vector varVector = (Vector) (variables.elementAt(subjectIndex));
		int valueIndex = varVector.indexOf(variable);
		if (valueIndex != -1) {
			return (String) (valVector.elementAt(valueIndex));
		}
		return "";
	}

	public void deleteValue(String subject, String variable) {
		int subjectIndex = subjects.indexOf(subject);
		if (subjectIndex == -1)
			return;

		Vector valVector = (Vector) (values.elementAt(subjectIndex));
		Vector varVector = (Vector) (variables.elementAt(subjectIndex));

		int valueIndex = varVector.indexOf(variable);
		if (valueIndex != -1) {

			valVector.removeElementAt(valueIndex);
			varVector.removeElementAt(valueIndex);

			int assignmentLine = findAssignmentLine(subject, variable);
			if (assignmentLine != -1) {
				lines.removeElementAt(assignmentLine);
			}

			if (varVector.size() == 0) {
				deleteSubject(subject);
			}
			if (saveOnChange)
				saveFile();
		}
	}

	public void deleteSubject(String subject) {
		int subjectIndex = subjects.indexOf(subject);
		if (subjectIndex == -1)
			return;

		values.removeElementAt(subjectIndex);
		variables.removeElementAt(subjectIndex);
		subjects.removeElementAt(subjectIndex);

		int start = findSubjectLine(subject);
		int end = endOfSubject(start);
		for (int i = start; i < end; i++) {
			lines.removeElementAt(start);
		}
		if (saveOnChange)
			saveFile();
	}

	public void saveFile() {
		try {
			DataOutputStream outFile = new DataOutputStream(
					new BufferedOutputStream(new FileOutputStream(fileName)));
			for (int i = 0; i < lines.size(); i++) {
				outFile.writeBytes((String) (lines.elementAt(i))
						+ System.getProperty("line.separator"));
			}
			outFile.close();
		} catch (IOException e) {
		}
	}

	protected void finalize() {
		saveFile();
	}

	public static void main(String[] args) {
		if (args.length != 1) {
			System.out.println("IniFile <filename>");
			System.exit(0);
		} else {
			String[] vars;
			IniFile f = new IniFile(args[0]);
			String[] subs = f.getSubjects();
			for (int i = 0; i < subs.length; i++) {
				System.out.println(subs[i]);
				vars = f.getVariables(subs[i]);
				for (int j = 0; j < vars.length; j++) {
					System.out.println("  '" + vars[j] + "' = '"
							+ f.getValue(subs[i], vars[j]) + "'");
				}
			}
		}
	}
}